from collections import defaultdict



class NameManager:

	"""
	The Namemanager is a attribute of the root Thing in a kinematic tree. It resolves name conflicts
	by finding synonymous Things and renaming them by adding a numeration. If a renamin results in a
	new naming conflicts, another round of renaming is applied unti all conflicts are resolved.
	
	For example, if there are four Geoms in a kinematic tree that are named by the user respectively:
	
	.. code-block::
		:caption: USERNAMES
				
		"rock", "rock", "rock_(1)", "stone"
	
	A name conflict exists for the first two geoms. The name conflict resolving after the first round 
	would result in the names

	.. code-block::
		:caption: RENAMING (1)

		"rock_(0)", "rock_(1)", "rock_(0)", "stone"
	
	Now a name conflict for the first and the third name exists, because the user name for the third 
	Geom coincidentally alligns with the numeration scheme of the resolver. Therefore a second renaming 
	round results in the names

	.. code-block::
		:caption: RENAMING (2)

		"rock_(0)_(0)", "rock_(1)", "rock_(0)_(1)", "stone"
	
	The cascading of numerations can be avoided by not ending user defined names with ``"_(<int>)"``.
	
	Attributes
	----------
	root : ThingType
		The root node for which the NameManager resolves nameing conflicts.
	type_scopes : dict
		Identical names only result in conflicts if they are given to Thing of the same type. The type_scopes dict splits the name resolving for all objects of one type.
	"""
	
	def __init__(self, root):
		"""
		The NameManager resolves naming conflicts for predefined names from the user and naming conflicts resulting from renaming of specific objects.
		
		Parameters
		----------
		root : ThingType
			A thing which is the root if its kinematic tree.
		"""
		self.root = root
		self.type_scopes = {}
		self.register()


	def __getitem__(self, key):
		"""
		Returns the type scope for a given Type.
		
		Parameters
		----------
		key : class
			The class for which the TypeScope is requested.
		
		Returns
		-------
		TypeScope
			The TypeScope that handles the name conflict resolving for all descendants of root of the given type.
		"""
		return self.type_scopes[key]


	def __setitem__(self, key, type_scope):
		"""
		Sets the TypeScope for the a given key Type.
		
		Parameters
		----------
		key : class
			The Type of objects for which the TypeScope is set.
		type_scope : TypeScope
			The TypeScope assigned to the NameManager.
		"""
		self.type_scopes[key] = type_scope


	def register(self):
		"""
		The calling of this method resolves naming conflicts.
		"""
		DESCENDANTS = defaultdict(list)
		for _, dec_dict in self.root.descendants.items():
			element_type = dec_dict['type']
			descendants  = dec_dict['descendants']
			DESCENDANTS[element_type].extend(descendants)
		#for name, dec_dict in self.root.descendants.items():
		for element_type, descendants in DESCENDANTS.items():
			descendants = list(dict.fromkeys(descendants))
			#element_type = dec_dict['type']
			#descendants  = dec_dict['descendants']
			self[element_type] = TypeScope(self, descendants)


	def unregister(self):
		"""
		The calling of this method reverses the effect of registering the NameManager.
		"""
		for type_scope in self.type_scopes.values():
			type_scope.unregister()
		self.type_scopes = {}


class TypeScope:

	"""
	This class upon initialization resolves name conflicts for all elements of a specifig Type.
	After renaming of assigned Things by the user, the `register` method is called automatically 
	to resolve potentially new conflicts.
	
	Attributes
	----------
	descendants : list
		A list of all Things that are captured under the TypeScope.
	name_manager : NameManager
		The NameManager to which the TypeScope is assigned.
	name_scopes : dict
		A dictionary containing the NameScopes for each user defined name.
	"""
	
	def __init__(self, name_manager, descendants):
		"""
		Upon initialization name conflicts are resolved for all descendants.
		
		Parameters
		----------
		name_manager : NameManager
			The NameManager to which the TypeScope is assigned.
		descendants : list
			A list of all Things that are captured under the TypeScope.
		"""
		self.name_manager = name_manager
		self.descendants  = descendants
		self.name_scopes  = {}
		self.register()


	def __getitem__(self, name: str):
		"""
		Returns the NameScope for a given name.
		
		Parameters
		----------
		name : str
			The name of the requested NameScope
		
		Returns
		-------
		NameScope
			The requested NameScope contains all Things decending from root, that are of the same Type and share a userdefined name.
		"""
		return self.name_scopes[name]


	def __setitem__(self, name: str, name_scope) -> None:
		"""
		Assigns a NameScope for a given name.
		
		Parameters
		----------
		name : str
			The name for which the NameScope is assigned.
		name_scope : NameScope
			The NameScope which is assigned to the given name.
		"""
		self.name_scopes[name] = name_scope


	def register(self):
		"""
		Calling this method resolves name conflicts for all assigned descendants of root. 
		After a assigned Thing is renamed register is called automatically to resolve 
		potential new conflicts.
		"""
		for descendant in self.descendants:
			descendant._name_scope = None
		self.name_scopes  = {name: NameScope(type_scope=self, 
						     name=name, 
						     descendants=descendants) for name, descendants in self.synonyms().items()}
		while True:
			synonyms = self.synonyms()
			if all(map(lambda x: len(x) == 1, synonyms.values())):
				break
			else:
				for name, descendants in synonyms.items():
					if len(descendants) > 1:
						self[name] = NameScope(self, name, descendants)


	def unregister(self):
		"""
		This method reverses the effect of registering for the TypeScope.
		"""
		for name_scope in self.name_scopes.values():
			name_scope.unregister()
		self.name_scopes = {}
		self.decendants  = {}


	def synonyms(self):
		"""
		This method gathers Things that are synomymous (share the same name). 
		This method takes renaming into acount, so after one round of renaming 
		has taken place, calling this method again, will result in synonyms 
		according to the previous renaming.
		
		Returns
		-------
		dict
			Keys of the dictionary are names and the values are lists containing all Things under the scope of the given name.
		"""
		synonyms = defaultdict(list)
		for descendant in self.descendants:
			synonyms[descendant.name].append(descendant)
		return synonyms



class NameScope:

	"""
	The NameScope class resolves the name conflicts for Things of the same Type 
	and the same assigned name (either by the user or by a previous name conflict 
	resolving iteration).
	
	Attributes
	----------
	descendants : list
	A list of all descendants of the same Type with the same name.
	type_scope : TypeScope
	The TypeScope to which the NameScope is assigned.
	"""
	
	def __init__(self, type_scope, name, descendants):
		"""
		After the initialization the name conflicts for all given devendents are resolved.
		
		Parameters
		----------
		type_scope : TypeScope
			The TypeScope to which the NameScope is assigned.
		name : str
			The initial name of all Things assigned to the NameScope.
		descendants : list
			A list of all descendants of the same Type with the same name.
		"""
		self.descendants = descendants
		self._name       = name
		self.type_scope  = type_scope
		self.register()


	def __len__(self):
		"""
		Gives the number of Things under this scope.
		
		Returns
		-------
		int
			The amount of assigned descendants.
		"""
		return len(self.descendants)


	def register(self):
		"""
		This method assignes all descendants to this NameScope. Afterwards 
		a unique name can be requested by the descendant.
		"""
		for descendant in self.descendants:
			descendant._name_scope = self


	def unregister(self):
		"""
		This method reverses the effect of registering all decendents of 
		the NameScope.
		"""
		for descendant in self.descendants:
			descendant._name_scope = None
		self.decendants = {}


	def name(self, descendant) -> str:
		"""
		Returns a unique name for a given descendant.
		
		Parameters
		----------
		descendant : Thing
			The Thing that is assigned to the NameScope.
		
		Returns
		-------
		str
			A unique name that is potentially altered by an enumeration scheme to resolve name conflicts.
		"""
		if len(self) == 1:
			return self._name
		else:
			idx = self.descendants.index(descendant)
			return f'{self._name}_({idx})'


	def __assign(self, descendant, name):
		"""
		This method assigns a name to a descendant.
		
		Parameters
		----------
		descendant : Thing
			A descendant to which a name is assigned.
		name : str
			The name to be assigned to the descendant.
		"""
		if name != self._name and name != descendant.name:
			descendant._name = name
			self.type_scope.register()



import os
import sys
import numpy as np
import xml.etree.ElementTree as xml
import blueprints as blue



class BaseGeom(blue.GeomType, blue.thing.NodeThing, blue.thing.MoveableThing, blue.thing.ColoredThing):

	"""
	Geoms introduce physical matter into the simulation. For different Geom shapes individual
	Geom classes are available, with shape depended size attributes. All other attributes are 
	specified in :class:`BaseGeom`. 

	.. note::
		Though units are not bound to one specific system of measurements, it is highly recommended 
		to use `MKS <https://en.wikipedia.org/wiki/MKS_system_of_units>`__, since all defaults are 
		defined based on meters, kilograms and seconds and most mujoco models online also use the 
		metric system. If you want to switch from the imperial to the metric system the following resource might be 
		of interest to `convert units from imperial to metric <https://xkcd.com/526/>`__. If you insist on 
		using imperial units, take a look at the following resource to `convert units from metric to imperial <https://rick.nerial.uk/>`__.
	
	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-geom>`__.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:                np.ndarray|list[int|float] = [0., 0., 0.],
		     alpha:              int|float = 0,
		     beta:               int|float = 0,
		     gamma:              int|float = 0, 
		     material:           blue.MaterialType|None = None, 
		     mass:               int|float|None = None,
		     density:            int|float      = 1000.,
		     margin:             int|float      = 0.0,
		     gap:                int|float      = 0.0,
		     sliding_friction:   int|float      = 1,
		     torsional_friction: int|float      = 0.005,
		     rolling_friction:   int|float      = 0.0001, 
		     shellinertia:       bool|None      = None,
		     color:              object|None    = None, 
		     name:               str|None       = None, 
		     x:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:                  int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:               list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		density : int | float | None, optional
			Material density used to compute the geom mass and inertia. The computation is based on 
			the geom shape and the assumption of uniform density. The internal default of 1000 is the 
			density of water in SI units. This attribute is used only when the mass attribute above is 
			unspecified.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		mass : int | float | None, optional
			If this attribute is specified, the density attribute below is ignored and the geom density
			is computed from the given mass, using the geom shape and the assumption of uniform density. 
			The computed density is then used to obtain the geom inertia. Recall that the geom mass and 
			inertia are only used during compilation, to infer the body mass and inertia if necessary.
		shellinertia : bool | None, optional
			If true, the geom’s inertia is computed assuming that all the mass is concentrated on the 
			boundary. In this case density is interpreted as surface density rather than volumetric 
			density.
		sliding_friction : int | float, optional
			Friction parameter for sliding used to compute the forces on contact pairs.
		torsional_friction : int | float, optional
			Friction parameter for torsion used to compute the forces on contact pairs.
		rolling_friction : int | float, optional
			Friction parameter for rolling used to compute the forces on contact pairs. 
		color : blue.ColorType
			The color of the Geom. For a detailed description see :class:`Color `
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.mass               = mass # float(mass) if mass is not None else mass
		self.density            = density
		self.shellinertia       = shellinertia
		self.margin             = margin
		self.gap                = gap
		self.sliding_friction   = sliding_friction
		self.torsional_friction = torsional_friction
		self.rolling_friction   = rolling_friction
		# MATERIAL
		self.material           = material
		# PSEUDO CHILDREN
		self._tendons           = []
		self._CHILDREN 		= {'tendons': {'type':     blue.TendonType, 
						       'children': self._tendons}}
		# SETTING COLOR BLUEPRINT COLOR DEFAULT FOR GEOMS
		color = color if color is not None else 'grey'
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 name=name, 
				 color=color, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@blue.restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		indicies : dict
			All indecies used for retrieving data during the simulation.
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		self._index       = indicies['geom']
		indicies['geom'] += 1
		if self.material is not None:
			self.material._build(parent, world, indicies, **kwargs)
			kwargs['material'] = self.material.asset.name
		return super()._build(parent, world, indicies, **kwargs)


	@blue.restrict
	@classmethod
	def _from_xml_element(cls,
			      xml_element: xml.Element,
			      material:    blue.MaterialType|None = None,
			      **kwargs) -> blue.ThingType:
		# Strip material name from XML before parsing — we inject the object directly
		mat_attr = xml_element.get('material')
		if mat_attr is not None:
			xml_element = xml.Element(xml_element.tag, {k: v for k, v in xml_element.items() if k != 'material'})
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['copy'] = False
		if material is not None:
			init_args['material'] = material
		init_args.update(kwargs)
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		return obj


	@property
	def type(self) -> str:
		"""
		This is a derived attribute, which is used to specify the ``type`` attribute in the 
		mujoco ``geom`` tag.

		Returns
		-------
		str
			The type is the lower case name of the Geom class.
		"""
		return self.__class__.__name__.lower()


	@property
	def material(self):
		"""
		Materials can be used to specify reflective properties of a Geom, 
		as well as to apply :class:`Textures <blueprints.texture.BaseTexture>` to them.

		Returns
		-------
		blue.MaterialType
			The Material of the Geom
		"""
		return self._material


	@material.setter
	@blue.restrict
	def material(self, material: blue.MaterialType|None) -> None:
		self._material = material.copy() if material is not None else material


	@property
	def friction(self) -> np.ndarray:
		"""
		Contact friction parameters for dynamically generated contact pairs. The first number is 
		the sliding friction, acting along both axes of the tangent plane. The second number is the 
		torsional friction, acting around the contact normal. The third number is the rolling 
		friction, acting around both axes of the tangent plane.

		Returns
		-------
		np.ndarray
			Individual components are found in :attr:`sliding_friction`, :attr:`torsional_friction` 
			and :attr:`rolling_friction`.
		"""
		return np.array([self.sliding_friction, 
				 self.torsional_friction, 
				 self.rolling_friction], dtype=np.float32)


	@friction.setter
	@blue.restrict
	def friction(self, friction: np.ndarray|list[int|float]):
		"""
		Contact friction parameters for dynamically generated contact pairs. The first number is 
		the sliding friction, acting along both axes of the tangent plane. The second number is the 
		torsional friction, acting around the contact normal. The third number is the rolling 
		friction, acting around both axes of the tangent plane.

		Parameters
		----------
		friction : np.ndarray | list[int | float]
			Individual components are found in :attr:`sliding_friction`, :attr:`torsional_friction` 
			and :attr:`rolling_friction`.
		"""
		self.sliding_friction   = friction[0]
		self.torsional_friction = friction[1]
		self.rolling_friction   = friction[2]


	@property
	def mass(self) -> float:
		"""
		If this attribute is specified, the density attribute below is ignored and the geom density
		is computed from the given mass, using the geom shape and the assumption of uniform density. 
		The computed density is then used to obtain the geom inertia. Recall that the geom mass and 
		inertia are only used during compilation, to infer the body mass and inertia if necessary.

		Unit defaults are set in SI units, but the physical properties can in principle be 
		interpreted in any system of measurements.

		Returns
		-------
		float
			The default unit convention is kilogram.
		"""
		return self._mass


	@mass.setter
	@blue.restrict
	def mass(self, mass: int|float|None) -> None:
		"""
		If this attribute is specified, the density attribute below is ignored and the geom density
		is computed from the given mass, using the geom shape and the assumption of uniform density. 
		The computed density is then used to obtain the geom inertia. Recall that the geom mass and 
		inertia are only used during compilation, to infer the body mass and inertia if necessary.

		Unit defaults are set in SI units, but the physical properties can in principle be 
		interpreted in any system of measurements.

		Parameters
		----------
		mass : int | float | None
			The default unit convention is kilogram.
		"""
		self._mass = float(mass) if mass is not None else self._DEFAULT_VALS()['mass']


	@property
	def density(self) -> float:
		"""
		Unit defaults are set in SI units, but the physical properties can in principle be 
		interpreted in any system of measurements.

		Material density used to compute the geom mass and inertia. The computation is based on 
		the geom shape and the assumption of uniform density. The internal default of 1000 is the 
		density of water in SI units. This attribute is used only when the mass attribute above is 
		unspecified.
		
		Returns
		-------
		float
			Density is either volumetric if :attr:`shellinertia` is false, otherwise its planar.
		"""
		return self._density


	@density.setter
	@blue.restrict
	def density(self, density: int|float|None) -> None:
		"""
		Unit defaults are set in SI units, but the physical properties can in principle be 
		interpreted in any system of measurements.

		Material density used to compute the geom mass and inertia. The computation is based on 
		the geom shape and the assumption of uniform density. The internal default of 1000 is the 
		density of water in SI units. This attribute is used only when the mass attribute above is 
		unspecified.
		
		Parameters
		----------
		density : int | float | None
			Density is either volumetric if :attr:`shellinertia` is false, otherwise its planar.
		"""
		self._density = float(density)


	@property
	def shellinertia(self) -> bool:
		"""
		If shellinertia is True all mass will be distributed along the surface of the object. 
		Changing this attribute also changes the effect of :attr:`density` from volumetric to planar.

		Returns
		-------
		bool
			Indicates whether mass is concentrated on the surface.
		"""
		return self._shellinertia


	@shellinertia.setter
	@blue.restrict
	def shellinertia(self, shellinertia: bool|None) -> None:
		"""
		If shellinertia is True all mass will be distributed along the surface of the object. 
		Changing this attribute also changes the effect of :attr:`density` from volumetric to planar.
		
		Parameters
		----------
		shellinertia : bool | None
			Indicates whether mass is concentrated on the surface.
		"""
		self._shellinertia = shellinertia


	@property
	def margin(self) -> float:
		"""
		Margin defines the minimum distance used to detect a contact.

		Returns
		-------
		float
			Default units are interpreted in meters.
		"""
		return self._margin


	@margin.setter
	@blue.restrict
	def margin(self, margin: int|float) -> None:
		"""
		Margin defines the minimum distance used to detect a contact.

		Parameters
		----------
		margin : int | float
			Default units are interpreted in meters.
		"""
		self._margin = float(margin)

	@property
	def gap(self) -> float:
		"""
		Gap defines the gap on minimum distance used to detect a contact.

		Returns
		-------
		float
			Default units are interpreted in meters.
		"""
		return self._gap


	@gap.setter
	@blue.restrict
	def gap(self, gap: int|float) -> None:
		"""
		Gap defines the gap on minimum distance used to detect a contact.

		Parameters
		----------
		gap : int | float
			Default units are interpreted in meters.
		"""
		self._gap = float(gap)


	@property
	def sliding_friction(self) -> float:
		"""
		Friction parameter for sliding used to computed the forces on contact pairs. 

		Returns
		-------
		float
			First component of :attr:`friction`.
		"""
		return self._sliding_friction


	@sliding_friction.setter
	@blue.restrict
	def sliding_friction(self, sliding_friction: int|float) -> None:
		"""
		Friction parameter for sliding used to computed the forces on contact pairs. 

		Parameters
		----------
		sliding_friction : int | float
			First component of :attr:`friction`.
		"""
		self._sliding_friction = float(sliding_friction)


	@property
	def torsional_friction(self) -> int|float:
		"""
		Friction parameter for torsion used to computed the forces on contact pairs. 

		Returns
		-------
		float
			Second component of :attr:`friction`.
		"""
		return self._torsional_friction


	@torsional_friction.setter
	@blue.restrict
	def torsional_friction(self, torsional_friction: int|float) -> None:
		"""
		Friction parameter for torsion used to computed the forces on contact pairs. 

		Parameters
		----------
		torsional_friction : int | float
			Second component of :attr:`friction`.
		"""
		self._torsional_friction = float(torsional_friction)


	@property
	def rolling_friction(self) -> int|float:
		"""
		Friction parameter for rolling used to computed the forces on contact pairs. 

		Returns
		-------
		float
			Third component of :attr:`friction`.
		"""
		return self._rolling_friction


	@rolling_friction.setter
	@blue.restrict
	def rolling_friction(self, rolling_friction: int|float) -> None:
		"""
		Friction parameter for rolling used to computed the forces on contact pairs. 

		Parameters
		----------
		rolling_friction : int | float
			Third component of :attr:`friction`.
		"""
		self._rolling_friction = float(rolling_friction)



class Capsule(blue.CapsuleGeomType, blue.tube.BaseTube, BaseGeom):

	"""
	Capsules consist of a cylinder with two half spheres at its ends. It can either be constructed from 
	its standard constructor for position, orientation, radius and length, or with the :meth:`blueprints.tube.BaseTube` 
	constructor. For a detailed description see :class:`blueprints.tube.BaseTube`.

	.. code-block:: python
		:caption: Standard constructor:

		>>> geom = blue.geoms.Capsule(alpha=TAU/8, length=3)
		>>> geom.head
		array([ 0.          1.06066017 -1.06066017], dtype=np.float32)
		>>> geom.tail
		array([ 0.         -1.06066017  1.06066017], dtype=np.float32)

	.. code-block:: python
		:caption: From points constructor:

		>>> geom = blue.geoms.Capsule.from_points([0, -1, 4], [2, 1, 0], radius=0.5)
		>>> geom.pos
		array([1. 0. 2.], dtype=np.float32)
		>>> geom.alpha, geom.beta, geom.gamma
		-2.6779450445889874 0.420534335283965 0.6847192030022825
		>>> geom.length
		4.898979485566356
	
	Attributes
	----------
	head : np.ndarra
		The first end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``head`` is the first argument given.
	tail : np.ndarra
		The second end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``tail`` is the second argument given.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:      np.ndarray|list[int|float] = [0., 0., 0.],
		     radius:   int|float   = 1.,
		     length:   int|float   = 1.,
		     alpha:    int|float   = 0.,
		     beta:     int|float   = 0.,
		     gamma:    int|float   = 0., 
		     margin:   int|float   = 0.0,
		     gap:      int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:    object|None = None, 
		     name:     str|None    = None, 
		     x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:     list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		radius : int | float, optional
			Represents the radius of the cylinder and the two half spheres.
		length : int | float, optional
			The length of the cylinder.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.radius = float(radius)
		self.length = float(length)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 material=material, 
				 margin=margin, 
				 gap=gap, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The size defines the radius and half length of the capsule.
		"""
		return np.array([self.radius, 
				 self.length/2], dtype=np.float32)


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			The size defines the radius and half length of the capsule.
		"""
		self.radius = float(size[0])
		self.length = float(size[1]) * 2



class Cylinder(blue.CylinderGeomType, blue.tube.BaseTube, BaseGeom):

	"""
	Cylinders can either be constructed from their standard constructor for position, orientation, 
	radius and length, or with the :meth:`blueprints.tube.BaseTube` constructor. For a detailed description 
	see :class:`blueprints.tube.BaseTube`.

	.. code-block:: python
		:caption: Standard constructor:

		>>> geom = blue.geoms.Cylinder(alpha=TAU/8, length=3)
		>>> geom.head
		array([ 0.          1.06066017 -1.06066017], dtype=np.float32)
		>>> geom.tail
		array([ 0.         -1.06066017  1.06066017], dtype=np.float32)

	.. code-block:: python
		:caption: From points constructor:

		>>> geom = blue.geoms.Cylinder.from_points([0, -1, 4], [2, 1, 0], radius=0.5)
		>>> geom.pos
		array([1. 0. 2.], dtype=np.float32)
		>>> geom.alpha, geom.beta, geom.gamma
		-2.6779450445889874 0.420534335283965 0.6847192030022825
		>>> geom.length
		4.898979485566356
	
	Attributes
	----------
	head : np.ndarra
		The first end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``head`` is the first argument given.
	tail : np.ndarra
		The second end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``tail`` is the second argument given.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:      np.ndarray|list[int|float] = [0., 0., 0.],
		     radius:   int|float   = 1.,
		     length:   int|float   = 1.,
		     alpha:    int|float   = 0.,
		     beta:     int|float   = 0.,
		     gamma:    int|float   = 0., 
		     margin:   int|float   = 0.0,
		     gap:      int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:    object|None = None, 
		     name:     str|None    = None, 
		     x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:     list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		radius : int | float, optional
			Represents the radius of the cylinder and the two half spheres.
		length : int | float, optional
			The length of the cylinder.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.radius = float(radius)
		self.length = float(length)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 margin=margin, 
				 gap=gap, 
				 material=material, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The size defines the radius and half length of the cylinder.
		"""
		return np.array([self.radius, 
				 self.length/2], dtype=np.float32)


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			 The size defines the radius and half length of the cylinder.
		"""
		self.radius = float(size[0])
		self.length = float(size[1]) * 2



class Box(blue.BoxGeomType, blue.tube.BaseTube, BaseGeom):

	"""
	Boxes can either be constructed from their standard constructor for position, orientation, 
	and x-, y-, z-length, or with the :meth:`blueprints.tube.BaseTube` constructor. For a detailed 
	description see :class:`blueprints.tube.BaseTube`.

	.. code-block:: python
		:caption: Standard constructor:

		>>> geom = blue.geoms.Box(x_length=2, y_length=4, z_length=1)
		>>> geom.head
		array([ 0.   0.  -0.5], dtype=np.float32)
		>>> geom.tail
		array([ 0.   0.   0.5], dtype=np.float32)

	.. code-block:: python
		:caption: From points constructor:

		>>> geom = blue.geoms.Box.from_points([4, 0,-2], [2, 4, 0], radius=0.5)
		>>> geom.pos
		array([ 3.   2.  -1.], dtype=np.float32)
		>>> geom.alpha, geom.beta, geom.gamma
		-1.1071487177940904 -0.4205343352839653 0.20135792079032988
		>>> geom.x_length, geom.y_length, geom.z_length
		0.5 0.5 4.898979485566356
	
	Attributes
	----------
	head : np.ndarra
		The first end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``head`` is the first argument given.
	tail : np.ndarra
		The second end of this Geoms. If it was defined by the :meth:`blueprints.tube.BaseTube` constructor, 
		``tail`` is the second argument given.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:      np.ndarray|list[int|float] = [0., 0., 0.], 
		     x_length: int|float   = 1., 
		     y_length: int|float   = 1., 
		     z_length: int|float   = 1., 
		     alpha:    int|float   = 0., 
		     beta:     int|float   = 0., 
		     gamma:    int|float   = 0.,  
		     margin:   int|float   = 0.0,
		     gap:      int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:    object|None = None, 
		     name:     str|None    = None, 
		     x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:     list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		x_length : int | float, optional
			The true length of the Box along the x-axis.
		y_length : int | float, optional
			The true length of the Box along the y-axis.
		z_length : int | float, optional
			The true length of the Box along the z-axis.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.x_length = float(x_length)
		self.y_length = float(y_length)
		self.z_length = float(z_length)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 margin=margin, 
				 gap=gap, 
				 material=material, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The size defines the half lengths of the Box.
		"""
		return np.array([self.x_length/2, 
				 self.y_length/2, 
				 self.z_length/2], dtype=np.float32)


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			The size defines the half lengths of the Box.
		"""
		self.x_length = float(size[0]) * 2
		self.y_length = float(size[1]) * 2
		self.z_length = float(size[2]) * 2



class Plane(blue.PlaneGeomType, BaseGeom):

	"""
	Planes can either be finite or infinite. They are normal to the Z-axis of their frame of reference 
	and the outside of the plane lies in the positive Z-axis. The inside of the plane is rendered semi 
	transparent. Finite Planes are rendered as rectangles with size specified in :attr:`x_length` 
	:attr:`y_length`. If those size parameters are set to zero or are set negative, the plain is instead 
	infinite. The :attr:`spacing` defines the size of grid subdivisions which are used in rendering.
	"""
	
	@blue.restrict
	def __init__(self,
		     pos:      np.ndarray|list[int|float] = [0., 0., 0.], 
		     x_length: int|float   = 0., 
		     y_length: int|float   = 0., 
		     spacing:  int|float   = 1., 
		     alpha:    int|float   = 0., 
		     beta:     int|float   = 0., 
		     gamma:    int|float   = 0., 
		     margin:   int|float   = 0.0,
		     gap:      int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:    object|None = None, 
		     name:     str|None    = None, 
		     x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:     list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		x_length : int | float, optional
			Length of the X-axis side. If both lengths are set to zero or negative the Plane is infinite.
		y_length : int | float, optional
			Length of the Y-axis side. If both lengths are set to zero or negative the Plane is infinite.
		spacing : int | float, optional
			Spacing between the grid subdivisions.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.x_length = float(x_length)
		self.y_length = float(y_length)
		self.spacing  = float(spacing)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 margin=margin, 
				 gap=gap, 
				 material=material, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The first two components are half lengths for the X-axis and the Y-axis and the third is the spacing between grid subdivisions.
		"""
		return np.array([self.x_length/2, 
				 self.y_length/2, 
				 self.spacing], dtype=np.float32)


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			The first two components are half lengths for the X-axis and the Y-axis and the third is the spacing between grid subdivisions.
		"""
		self.x_length = float(size[0]) * 2
		self.y_length = float(size[1]) * 2
		self.spacing  = float(size[2])



class Sphere(blue.SphereGeomType, BaseGeom):

	"""
	Spheres are defined via radius and position.
	
	Attributes
	----------
	radius : float
		The radius of the sphere
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:    np.ndarray|list[int|float] = [0., 0., 0.],
		     radius: int|float   = 1.,
		     alpha:  int|float   = 0.,
		     beta:   int|float   = 0.,
		     gamma:  int|float   = 0., 
		     margin: int|float   = 0.0,
		     gap:    int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:  object|None = None, 
		     name:   str|None    = None, 
		     x:      int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:      int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:      int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:   list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the Thing. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		radius : int | float, optional
			The radius of the Sphere.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.radius = float(radius)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 margin=margin, 
				 gap=gap, 
				 material=material, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The only component of size is the radius, which is interpreted as meters by default parameters and convention.
		"""
		return np.array([self.radius])


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			The only component of size is the radius, which is interpreted as meters by default parameters and convention.
		"""
		self.radius = float(size[0])



class Ellipsoid(blue.EllipsoidGeomType, BaseGeom):

	"""
	Ellipsoids are defined with via :attr:´x_radius`, :attr:´y_radius` and :attr:´z_radius`. The surface of 
	the Ellipsoid contains the solutions to the equation:

	:math:`\\left( \\frac{x}{r_x} \\right)^2 + \\left( \\frac{y}{r_y} \\right)^2 + \\left( \\frac{z}{r_z} \\right)^2 = 1`
	
	Attributes
	----------
	x_radius : float
		This attribute defines the radius for the X-axis.
	y_radius : float
		This attribute defines the radius for the Y-axis.
	z_radius : float
		This attribute defines the radius for the Z-axis.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:      np.ndarray|list[int|float] = [0., 0., 0.],
		     x_radius: int|float   = 1.,
		     y_radius: int|float   = 2.,
		     z_radius: int|float   = 3.,
		     alpha:    int|float   = 0.,
		     beta:     int|float   = 0.,
		     gamma:    int|float   = 0., 
		     margin:   int|float   = 0.0,
		     gap:      int|float   = 0.0,
		     material: blue.MaterialType|None = None, 
		     color:    object|None = None, 
		     name:     str|None    = None, 
		     x:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:        int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:     list[int|float]|np.ndarray|None = None, 
		     **kwargs):
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		x_radius : int | float, optional
			This attribute defines the radius for the X-axis.
		y_radius : int | float, optional
			This attribute defines the radius for the Y-axis.
		z_radius : int | float, optional
			This attribute defines the radius for the Z-axis.
		alpha : int | float | None, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also changes the :attr:`euler` property.
		beta : int | float | None, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also changes the :attr:`euler` property.
		gamma : int | float | None, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also changes the :attr:`euler` property.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		color : object
			The color of the Geom. See :class:`Color <blueprints.thing.colored.Color>` for a detailed description.
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		self.x_radius = float(x_radius)
		self.y_radius = float(y_radius)
		self.z_radius = float(z_radius)
		super().__init__(pos=pos, 
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 margin=margin, 
				 gap=gap, 
				 material=material, 
				 color=color, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 quat=quat, 
				 **kwargs)


	@property
	def size(self) -> np.ndarray:
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Returns
		-------
		np.ndarray
			The components contain the :attr:`x_radius`, :attr:`y_radius` and :attr:`z_radius` attribute.
		"""
		return np.array([self.x_radius, 
				 self.y_radius, 
				 self.z_radius], dtype=np.float32)


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[int|float]):
		"""
		.. note::
			In mujoco half lengths are used instead of true lengths of objects. This makes distance 
			calculations easier, but we find that it is overall more confusing then beneficial which
			is why blueprints uses proper lengths. Conversion is handled in the background, such that 
			users only need to specify true lengths.
		
		Parameters
		----------
		size : np.ndarray | list[int | float]
			The components contain the :attr:`x_radius`, :attr:`y_radius` and :attr:`z_radius` attribute.
		"""
		self.x_radius = float(size[0])
		self.y_radius = float(size[1])
		self.z_radius = float(size[2])



class Mesh(blue.MeshGeomType, BaseGeom):

	"""
	Meshes can be used to build more complex shapes either by specifying vertecies manually or by loading 
	mesh files directly.

	>>> blue.geoms.Mesh(filename='utahteapot.stl')
	Mesh<unnamed_mesh:1>
	>>> vertecies = [[0, 0, 0], 
		         [0, 2, 0], 
		         [2, 2, 0], 
		         [2, 0, 0], 
		         [1, 1, 2]]
	>>> mesh = blue.geoms.Mesh(vertecies=vertecies, centered=True)
	>>> mesh.vertecies
	[[-1. -1. -1.]
	 [-1.  1. -1.]
	 [ 1.  1. -1.]
	 [ 1. -1. -1.]
	 [ 0.  0.  1.]]


	The vertex data is accessed through a :class:`blueprints.assets.MeshAsset` which 
	in turn stores the data in a :class:`blueprints.cache.MeshCache`. As long as no vertex data are 
	modified copies of the :class:`Mesh` will reference the same Asset to avoid redundancy, otherwise a 
	new Asset and or Cache is created. The creation of a new Asset is computationally fairly cheap, the 
	creation of a new Cache may come with significant costs if the mesh file is large.

	Modifications of the following attributes might trigger the creation of a new Asset:

	1. ``pos``
	2. ``euler``
	3. ``scale``

	Modifications of the following attributes might trigger the creation of a new Cache:

	1. ``size``
	2. ``vertecies``


	Raises
	------
	Exception
		If neither a filename, vertecies or an asset are given an error is raised.
	"""

	@blue.restrict
	def __init__(self, 
		     pos:       np.ndarray|list[int|float]            = [0., 0., 0.],
		     vertecies: list[np.ndarray|list[float|int]]|None = None,  
		     filename:  str|None                              = None, 
		     centered:  bool                                  = True, 
		     asset:     blue.assets.MeshAsset|None            = None, 
		     material:  blue.MaterialType|None                = None, 
		     margin:    int|float   = 0.0,
		     gap:       int|float   = 0.0,
		     x:         int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:         int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:         int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the object. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		vertecies : list[np.ndarray | list[float | int]] | None, optional
			Either a list of positions of vertecies or None.
		filename : str | None, optional
			The name for the file from which the Mesh data is loaded. The supported formats are ``.stl`` in binary and ``.obj`` in binary and ascii.
		centered : bool, optional
			If this argument is set, vertecies are centered around the references frames origin.
		asset : blue.assets.MeshAsset | None, optional
			A possible asset, from which the Mesh is constructed.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		
		Raises
		------
		Exception
			If neither a filename, vertecies or an asset are given an error is raised.
		"""
		kwargs_nameless = kwargs.copy()
		if 'name' in kwargs_nameless:
			del kwargs_nameless['name']
		if sum(x is not None for x in (vertecies, filename, asset)) > 1:
			raise Exception('Not more than one argument (vertecies, filename or asset) is allowed to be None.')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		if asset is not None:
			self.asset = asset
		elif filename is not None:
			if not os.path.isfile(filename):
				# TRY DIRNAME PREFIX TO RESOLVE RELATIVE REF TO MAIN.PY
				filename = f'{os.path.dirname(sys.argv[0])}/{filename}'
				if not os.path.isfile(filename):
					raise Exception(f'File not found for path {filename}')
			if not os.path.isabs(filename):
				filename = os.path.abspath(filename)
			if not os.path.isabs(filename):
				path = os.path.abspath(os.path.dirname(sys.argv[0]))
				filename = f'{path}/{filename}'
			self.asset = blue.assets.MeshAsset(filename=filename, 
							   pos=pos, 
							   centered=centered, 
							   xml_data=vertecies is not None, 
							   **kwargs_nameless)
		elif vertecies is not None:
			self.asset = blue.assets.MeshAsset(vertecies=vertecies, 
							   pos=pos, 
							   centered=centered, 
							   xml_data=True, 
							   **kwargs_nameless)
		else:
			raise Exception('No asset could be constructed! Please provide either a filename, vertcies or an asset.')
		# SEPARATE CHANGES FOR POST-FREEZE ASSIGNMENT
		self.asset.freeze = True
		kwchanges   = {}
		if len(self.asset._references) > 0:
			proto_parent = next(iter(self.asset._references))
			for key, val in kwargs.items():
				if hasattr(proto_parent, key):
					equal = getattr(proto_parent, key) == val
					equal = bool(np.all(equal)) if isinstance(equal, np.ndarray) else equal
					if not equal:
						kwchanges[key] = val
						kwargs[key] = getattr(proto_parent, key)
			if not bool(np.all(proto_parent.pos == pos)):
				kwchanges['pos'] = pos
				pos = proto_parent.pos
		self.asset._add(self)
		#print(self.copy)
		super().__init__(pos=pos, 
				 material=material, 
				 margin=margin, 
				 gap=gap, 
				 **kwargs)
		self.asset.freeze = False
		for key, val in kwchanges.items():
			if key in self._BLUEPRINT_ATTR():
				setattr(self, key, val)


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		if self.material is not None:
			self.material._build(parent, world, indicies, **kwargs)
			kwargs['material'] = self.material.asset.name
		self._xml_root = xml.SubElement(parent, 
						self._MUJOCO_OBJ, 
						mesh=self.asset.name, 
						**self._mujoco_specs(kwargs))
		if self.asset._built:
			self._index = self.asset._index
		else:
			self._index       = indicies['mesh']
			indicies['mesh'] += 1
		self.asset._build(parent=parent, 
				  world=world, 
				  indicies=indicies, 
				  **kwargs)
		return self._xml_root


	@blue.restrict
	@classmethod
	def _from_xml_element(cls,
			      xml_element: xml.Element,
			      asset:       blue.AssetType,
			      material:    blue.MaterialType|None = None) -> blue.ThingType:
		"""
		This method reconstructs a Mesh from an xml element.

		Parameters
		----------
		xml_element : xml.Element
			The xml element from which a Mesh is reconstructed.
		asset : blue.AssetType
			The asset from which the Mesh takes its data.
		material : blue.MaterialType | None, optional
			The Material assigned to the Mesh.

		Returns
		-------
		blue.ThingType
			The reconstructed Mesh.
		"""
		# Strip material name from XML before parsing
		if xml_element.get('material') is not None:
			xml_element = xml.Element(xml_element.tag, {k: v for k, v in xml_element.items() if k != 'material'})
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		geom_type = rest_args['type']
		geom = object.__new__(blue.REGISTER.GEOM_THINGS[geom_type])
		init_args['asset'] = asset
		if material is not None:
			init_args['material'] = material
		geom.__init__(**init_args)
		for key, val in post_args.items():
			setattr(geom, key, val)
		return geom


	@property
	def vertecies(self) -> list[np.ndarray]:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly. 
		
		Returns
		-------
		list[np.ndarray]
			A list of all vertex positions.
		"""
		return self.asset.vertecies.copy()


	@vertecies.setter
	@blue.restrict
	def vertecies(self, vertecies: np.ndarray|list[np.ndarray|list[float|int]]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly. 
		
		Parameters
		----------
		vertecies : np.ndarray | list[np.ndarray | list[float | int]]
			A list of all vertex positions.
		
		"""
		self.asset._prepare_for_modification(self)
		self.asset.vertecies = vertecies


	@property
	def pos(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Returns
		-------
		np.ndarray | list[float | int]
		"""
		return self.asset.pos.copy()


	@pos.setter
	@blue.restrict
	def pos(self, pos: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Parameters
		----------
		pos : np.ndarray | list[float | int]
			Description
		"""
		self.asset._prepare_for_modification(self)
		self.asset.pos = pos


	@property
	def _pos(self) -> np.ndarray:
		"""
		Returns
		-------
		np.ndarray | list[float | int]
		"""
		return self.asset._pos.copy()


	@_pos.setter
	@blue.restrict
	def _pos(self, pos: np.ndarray|list[float|int]) -> None:
		"""
		Parameters
		----------
		pos : np.ndarray | list[float | int]
		"""
		self.asset._prepare_for_modification(self)
		self.asset._pos = pos


	@property
	def euler(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Returns
		-------
		np.ndarray
		"""
		return self.asset.euler


	@euler.setter
	@blue.restrict
	def euler(self, euler: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Parameters
		----------
		euler : np.ndarray | list[float | int]
			Description
		
		Deleted Parameters
		------------------
		vertex : list[np.ndarray | list[float | int]]
		Description
		pos : np.ndarray | list[float | int]
		Description
		"""
		self.asset._prepare_for_modification(self)
		self.asset.euler = euler


	@property
	def size(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly. 
		
		.. note::
			Altering this attribute also changes the :attr:`vertecies` attribute. To get the same 
			effect without changing the vertecies, see :attr:`scale`.
		
		Returns
		-------
		np.ndarray
			The size of the Mesh is calculated as the size of the Box with edges along the axes of 
			the Meshes frame of reference, that captures all vertecies.
		"""
		return self.asset.size


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly.

		.. note::
			Altering this attribute also changes the :attr:`vertecies` attribute. To get the same 
			effect without changing the vertecies, see :attr:`scale`.
		
		Parameters
		----------
		size : np.ndarray | list[float | int]
			Setting the size of the Mesh is scales the vertecies along the axes of the Meshes frame 
			of reference such that it fits into a Box of the given size.
		"""
		self.asset._prepare_for_modification(self)
		self.asset.size = size


	@property
	def scale(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		.. note::
			If :attr:`vertecies` should be changed directly such that it adheres to the given scale, 
			use :attr:`size` instead.
		
		Returns
		-------
		np.ndarray
			The scaling of X-axis, Y-axis and Z-axis in each component. Analogous to :attr:`size` without changing :attr:`vertecies`.
		"""
		return self.asset.scale


	@scale.setter
	@blue.restrict
	def scale(self, scale: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		.. note::
			If :attr:`vertecies` should be changed directly such that it adheres to the given scale, 
			use :attr:`size` instead.
		
		Parameters
		----------
		scale : np.ndarray | list[float | int]
			The scaling of X-axis, Y-axis and Z-axis in each component. Analogous to :attr:`size` without changing :attr:`vertecies`.
		"""
		self.asset._prepare_for_modification(self)
		self.asset.scale = scale



class HField(blue.HFieldGeomType, BaseGeom):
	"""
	Height field are mesh-like Geoms with a rectangular shape and varing heights. 
	The height data can either be given as a file or as a 2D numpy array (alternatively 2D list). 
	Supported fileformats are PNG and the mujoco native HF format.
	
	.. code-block::
		:caption: Terrain Data
	
		>>> ...
		>>> spiral = np.array([ ... ])
		>>> hfield = blue.geoms.HField(terrain=spiral)
		# ALTERNATIVELY
		>>> hfield = blue.geoms.HField(filename='spiral.hf')
		>>> world.attach(hfield, plane, ball)

	.. image:: /_static/spiral.gif

	Terrain data can be edited at runtime by setting the hfield attribute.

	.. code-block::
		:caption: Runtime Modification

		>>> n_frames = 1000
		>>> terrain = blue.perlin((n_frames, 100, 100), periodic=True)
		>>> ... # FURTHER TERRAIN PROCESSING
		>>> hfield = blue.geoms.HField(terrain=terrain, color='orange')
		>>> world.attach(hfield, copy=False)
		>>> for n in range(total_steps//20):
		>>> 	    world.step(n_steps=20)
		>>> 	    hfield.terrain = terrain[n % n_frames, ...]

	.. image:: /_static/hfield_update.gif

	"""
	@blue.restrict
	def __init__(self, 
		     pos:           np.ndarray|list[int|float]      = [0., 0., 0.], 
		     terrain:       np.ndarray|list[int|float]|None = None, 
		     filename:      str|None                        = None, 
		     x_length:      int|float|None                  = 1, 
		     y_length:      int|float|None                  = 1, 
		     z_length:      int|float|None                  = 1, 
		     height_offset: int|float|None                  = 1, 
		     margin:        int|float   = 0.0,
		     gap:           int|float   = 0.0,
		     material:      blue.MaterialType|None          = None, 
		     asset:         blue.assets.HFieldAsset|None    = None, 
		     x:             int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:             int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:             int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		pos : list[int | float] | np.ndarray | None, optional
			Represents the position of the Thing. Changing this attribute also changes the properties :attr:`x`, :attr:`y` and :attr:`z`.
		terrain : np.ndarray | list [ int | float ] | None
			The terrain data in the shape of a 2D array.
		filename : str | None, optional
			The name for the file from which the HField data is loaded. The supported formats are ``.png`` and ``.hf``.
		x_length : int | float, optional
			The length of the HField along the x-axis.
		y_length : int | float, optional
			The length of the HField along the y-axis.
		z_length : int | float, optional
			The length of the HField along the z-axis.
		height_offset : int | float | None
			The offset in height for the terrain data points.
		margin : int | float | None, optional
			A contact is considered active only if the distance between the two geom surfaces is below 
			margin-gap.
		gap : int | float | None, optional
			This attribute is used to enable the generation of inactive contacts, i.e., contacts that are 
			ignored by the constraint solver but are included in mjData.contact for the purpose of custom 
			computations. When this value is positive, geom distances between margin and margin-gap 
			correspond to such inactive contacts.
		material : blue.MaterialType, optional
			The :class:`Material <blueprints.material.Material>` of the Geom. 
			:class:`Textures <blueprints.texture.BaseTexture> are applied via Materials.
		asset : blue.assets.HFieldAsset | None, optional
			The Asset of the HField
		name : str | None, optional
			The user specified name might potentially be altered to avoid a naming conflict by appending an enumeration scheme.
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		**kwargs
			Keyword arguments passed to ``super().__init__``
		"""
		#self.terrain = terrain
		kwargs_nameless = kwargs.copy()
		if 'name' in kwargs_nameless:
			del kwargs_nameless['name']
		if sum(x is not None for x in (terrain, filename, asset)) > 1:
			raise Exception('Not more than one argument (terrain, file or asset) is allowed to be None.')
		if pos is None:
			x = float(x) if x is not None else 0.
			y = float(y) if y is not None else 0.
			z = float(z) if z is not None else 0.
			pos = np.array([x, y, z], dtype=np.float32)
		if asset is not None:
			self.asset = asset
		elif filename is not None:
			if not os.path.isfile(filename):
				# TRY DIRNAME PREFIX TO RESOLVE RELATIVE REF TO MAIN.PY
				filename = f'{os.path.dirname(sys.argv[0])}/{filename}'
				if not os.path.isfile(filename):
					raise Exception(f'File not found for path {filename}')
			if not os.path.isabs(filename):
				filename = os.path.abspath(filename)
			if not os.path.isabs(filename):
				path = os.path.abspath(os.path.dirname(sys.argv[0]))
				filename = f'{path}/{filename}'
			self.asset = blue.assets.HFieldAsset(filename=filename, 
							     pos=pos, 
							     x_length=x_length, 
							     y_length=y_length, 
							     z_length=z_length, 
							     height_offset=height_offset, 
							     xml_data=False, 
							     **kwargs_nameless)
		elif terrain is not None:
			self.asset = blue.assets.HFieldAsset(terrain=terrain, 
							     pos=pos, 
							     x_length=x_length, 
							     y_length=y_length, 
							     z_length=z_length, 
							     height_offset=height_offset, 
							     xml_data=True, 
							     **kwargs_nameless)
		else:
			raise Exception('No asset could be constructed! Please provide either a filename, terrain or an asset.')
		self.asset._add(self)
		self.asset.freeze = True
		super().__init__(pos=pos,
				 material=material, 
				 margin=margin, 
				 gap=gap, 
				 **kwargs)
		self.asset.freeze = False


	def __getitem__(self, 
			key: tuple[slice|int]) -> np.ndarray|np.float32:
		"""
		Returns
		-------
		np.ndarray | np.float32
			The values from the selected index/slice of the height field.
		"""
		return self.asset[key]


	def __setitem__(self, 
			key:   tuple[slice|int], 
			value: int|float|list[int|float]|np.ndarray) -> None:
		"""
		Parameters
		----------
		key : tuple[slice | int]
			The indecies/slices of acces
		value : int|float|list[int|float]|np.ndarray
			The value to be assigned in the selected parts of the field.
		"""
		self.asset._prepare_for_modification(self)
		self.asset[key] = value


	@blue.restrict
	def _build(self, 
		   parent, 
		   world, 
		   indicies, 
		   **kwargs):
		"""
		This method is called to build the xml.
		
		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially
		
		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		if self.material is not None:
			self.material._build(parent, world, indicies, **kwargs)
			kwargs['material'] = self.material.asset.name
		self._xml_root = xml.SubElement(parent, 
						self._MUJOCO_OBJ, 
						hfield=self.asset.name, 
						**self._mujoco_specs(kwargs))
		if self.asset._built:
			self._index = self.asset._index
		else:
			self._index = indicies['hfield']
		self.asset._build(parent=parent, 
				  world=world, 
				  indicies=indicies, 
				  **kwargs)
		return self._xml_root

	# MUJOCO PROPERTIES

	@property
	def pos(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Returns
		-------
		np.ndarray | list[float | int]
		"""
		return self.asset.pos.copy()


	@pos.setter
	@blue.restrict
	def pos(self, pos: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Parameters
		----------
		pos : np.ndarray | list[float | int]
			Description
		
		Deleted Parameters
		------------------
		vertex : list[np.ndarray | list[float | int]]
		Description
		"""
		self.asset._prepare_for_modification(self)
		self.asset.pos = pos


	@property
	def terrain(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Returns
		-------
		np.ndarray | list[float | int]
		"""
		return self.asset.terrain.copy()


	@terrain.setter
	@blue.restrict
	def terrain(self, terrain: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Parameters
		----------
		terrain : np.ndarray | list[float | int]
			Description
		
		Deleted Parameters
		------------------
		vertex : list[np.ndarray | list[float | int]]
		Description
		"""
		#self.asset._prepare_for_modification(self)
		self.asset.terrain = terrain


	@property
	def euler(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Returns
		-------
		np.ndarray
		"""
		return self.asset.euler.copy()


	@euler.setter
	@blue.restrict
	def euler(self, euler: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.assets.MeshAsset` object if 
		the previous Asset was the children of multiple parents. Otherwise the Asset get modified 
		directly. 
		
		Parameters
		----------
		euler : np.ndarray | list[float | int]
			Description
		
		Deleted Parameters
		------------------
		vertex : list[np.ndarray | list[float | int]]
		Description
		pos : np.ndarray | list[float | int]
		Description
		"""
		self.asset._prepare_for_modification(self)
		self.asset.euler = euler


	@property
	def size(self) -> np.ndarray:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly. 
		
		.. note::
			Altering this attribute also changes the :attr:`vertecies` attribute. To get the same 
			effect without changing the vertecies, see :attr:`scale`.
		
		Returns
		-------
		np.ndarray
			The size of the Mesh is calculated as the size of the Box with edges along the axes of 
			the Meshes frame of reference, that captures all vertecies.
		"""
		return self.asset.size.copy()


	@size.setter
	@blue.restrict
	def size(self, size: np.ndarray|list[float|int]) -> None:
		"""
		Modifying this attribute will create a new :class:`blueprints.cache.MeshCache` object if 
		the previous Cache was the children of multiple parents. Otherwise the Cache get modified 
		directly.

		.. note::
			Altering this attribute also changes the :attr:`vertecies` attribute. To get the same 
			effect without changing the vertecies, see :attr:`scale`.
		
		Parameters
		----------
		size : np.ndarray | list[float | int]
			Setting the size of the Mesh is scales the vertecies along the axes of the Meshes frame 
			of reference such that it fits into a Box of the given size.
		"""
		self.asset._prepare_for_modification(self)
		self.asset.size = size

	# BLUEPRINTS ATTRIBUTES


	@property
	def x_length(self) -> float:
		"""
		Length along the ``X``-axis of the HField.

		Returns
		-------
		float
		"""
		return self.asset.x_length


	@property
	def y_length(self) -> float:
		"""
		Length along the ``Y``-axis of the HField.

		Returns
		-------
		float
		"""
		return self.asset.y_length


	@property
	def z_length(self) -> float:
		"""
		Length along the ``Z``-axis of the HField.

		Returns
		-------
		float
		"""
		return self.asset.z_length


	@property
	def height_offset(self):
		"""
		Height offset along the ``Z``-axis of the HField.

		Returns
		-------
		float
		"""
		return self.asset.height_offset


	@x_length.setter
	@blue.restrict
	def x_length(self, x_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self.asset.x_length = float(x_length)


	@y_length.setter
	@blue.restrict
	def y_length(self, y_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self.asset.y_length = float(y_length)


	@z_length.setter
	@blue.restrict
	def z_length(self, z_length: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self.asset.z_length = float(z_length)


	@height_offset.setter
	@blue.restrict
	def height_offset(self, height_offset: int|float|np.int32|np.int64|np.float32|np.float64) -> None:
		self.asset.height_offset = float(height_offset)
	




GEOM_THINGS = {'capsule':   Capsule, 
	       'cylinder':  Cylinder, 
	       'plane':     Plane, 
	       'sphere':    Sphere, 
	       'ellipsoid': Ellipsoid, 
	       'box':       Box, 
	       'mesh':      Mesh, 
	       'hfield':    HField}



"""
.. note::
	If you found yourself here not being a dev, you should probably :mod:`skip this section <blueprints.utils>` from or 
	look up the correspinging docs page of the Thing you were looking after for which this module only 
	implements its corresponding mirror types. A link to the docs page is to be found in the Types doc 
	string.
"""



import blueprints as blue
from blueprints import restrict
from . import base

import numpy as np
import inspect
import xml.etree.ElementTree as xml
from copy import copy
from collections import defaultdict



class UniqueThing(blue.UniqueThingType, base.BaseThing):

	"""
	Things that have multiple parents and that should not be copied to often can inherit from 
	:class:`UniqueThing`. If the attributes of this Thing should be altered for one parent but stay 
	the same for other parents, a copy is created which then can be altered. If the the Thing has only 
	one parent the attributes are instead altered directly. To do this, parents can call the 
	:meth:`_prepare_for_modification` method passing themselves which returns either a copy of the 
	Thing or the Thing itself which can then be altered.

	>>> alterable_child = node_thing.unique_thing._prepare_for_modification(node_thing)
	>>> atterable_child.attr = value

	This usecase is important if a Thing is modified often, but contains large amounts of data like 
	:class:`blueprints.cache.MeshCache` to avoid redundant copies.
	
	Parameters
	----------
	**kwargs
		Keyword arguments are passed to ``super().__init__``
		
	Attributes
	----------
	freeze : bool
		Setting this attribute freezes the setting of other attributes until it is unset.
	"""

	def __init__(self, **kwargs):
		"""
		Parameters
		----------
		**kwargs
			Keyword arguments are passed to ``super().__init__``
		"""
		self._references = set()
		self.freeze      = False
		super().__init__(**kwargs)
		assert hasattr(self, '_REFERENCE_NAME')


	def __setattr__(self, attr, val):
		"""
		Parameters
		----------
		attr : str
			The name of the attribute
		val : object
			The value assigned to the attribute
		"""
		if attr == 'freeze' or not hasattr(self, 'freeze') or not self.freeze:
			super().__setattr__(attr, val)


	def _remove(self, thing):
		"""
		Parameters
		----------
		thing : ThingType
			thing is detached as a parent of the :class:`UniqueThing`.
		"""
		if thing in self:
			self._references.remove(thing)


	def _add(self, thing):
		"""
		Parameters
		----------
		thing : ThingType
			thing is added as a parent of the :class:`UniqueThing`.
		"""
		if thing not in self:
			if hasattr(thing, self._REFERENCE_NAME):
				getattr(thing, self._REFERENCE_NAME)._remove(thing)
			self._references.add(thing)
			setattr(thing, self._REFERENCE_NAME, self)


	def _prepare_for_modification(self, parent):
		"""
		Parameters
		----------
		parent : ThingType
			The parent is given such that if a copy is generated if can be detached from the 
			:class:`UniqueThing` parents.
		"""
		assert getattr(parent, self._REFERENCE_NAME) is self
		assert parent in self
		#if self.parent is not None:
		#	self.root = self.parent.root
		if len(self._references) > 1 and not self.freeze:# and (not isinstance(self.root, blue.WorldType) or not self.root._is_built):
			thing = self.copy()
			thing._add(parent)


	def __contains__(self, thing):
		"""
		Parameters
		----------
		thing : ThingType
			A potential parent of this instance
		
		Returns
		-------
		bool
			Indicates whether thing is a parent of this instance.
		"""
		return thing in self._references



"""
Cameras provide a rendered image for each time step. A :class:`Camera` can be attached to a :class:`blueprints.body.Body` 
or to the :class:`blueprints.world.World` directly. If a Camera is attached to a :class:`blueprints.body.Body` 
in a :class:`Agent <blueprints.agent.Abent>` its rendered images are contained in the Agents observations. All 
cameras can be viewed when :meth:`blueprints.world.World.view` by setting the ``camera`` in the ``Rendering`` 
section.

Camera Orientation
------------------

Similarly to :class:`blueprints.light.Light` Cameras can either be fixed in orientation locally/globally or 
tracking another Body by setting the :attr:`Camera.mode` and optionally :attr:`Camera.target`. For a detailed 
description see :attr:`Camera.mode`.

.. code-block::
	:caption: Local orientation

	>>> body = blue.Body(cameras=blue.Camera(mode='fixed'))

.. code-block::
	:caption: Global orientation

	>>> body = blue.Body(cameras=blue.Camera(mode='track'))

.. code-block::
	:caption: Tracking Body

	>>> one  = blue.Body(name='one', cameras=blue.Camera(mode='targetbody'))
	>>> two  = blue.Body(name='two')
	>>> root = blue.Body(name='root', bodies=[one, two])
	>>> # Version one:
	>>> root.bodies['one'].cameras.target = root.bodies['two']
	>>> # Version two:
	>>> root.bodies['two'].targeting_cameras = root.bodies['one'].cameras

.. code-block:: mxml
	:caption: XML example
	
	<body name="root">
	    <body name="one">
	        <camera mode="targetbody" name="unnamed_camera" target="two" />
	    </body>
	    <body name="two" />
	</body>

Using the ``target`` attribute for a Camera results in continous tracking of the target Thing. 
If instead a Camera should be initially oriented to center an object, without tracking it during 
the simulation, the :meth:`Camera.looking` method can be used to align the camera to a position or 
direction.

.. code-block::
	:caption: Looking in a Direction

	>>> camera = camera.looking(direction=[1, 1, 1])

.. code-block:: mxml
	:caption: Looking in a Direction

	<camera euler="2.3561945 -0.6154797 0.5235988" />


To center a position either a Thing or a global position can be passed.


.. code-block::
	:caption: Looking at Positions/Things

	>>> body = blue.Body(pos=[-1, 0, 1])
	# VERSION 1
	>>> camera = camera.looking(pos=body)
	# VERSION 2
	>>> camera = camera.looking(pos=[-1, 0, 1])

.. code-block:: mxml
	:caption: Looking at Positions/Things

	<camera euler="1.5707964 2.3561945 1.5707964" />


Recording
---------
To record videos with a Camera you can use the methods :meth:`World.start_recording <blueprints.world.World.start_recording>`, 
:meth:`World.step <blueprints.world.World.step>` and :meth:`World.stop_recording <blueprints.world.World.stop_recording>`.
For more details refer to :meth:`start_recording <blueprints.world.World.start_recording>`.

.. code-block::
	:caption: Recording Videos

	>>> world.start_recording(cam_1, cam_2, cam_3, filename='my.gif', loop=0)
	>>> world.step(seconds=5)
	>>> world.stop_recording()
"""

import xml.etree.ElementTree as xml
import numpy as np
import mujoco
from mujoco.glfw import glfw

import blueprints as blue
from blueprints.utils.geometry import RADIANS_TO_DEGREES, DEGREES_TO_RADIANS, TAU
import time



class Camera(blue.CameraType, blue.thing.CyclicalThing, blue.thing.MoveableThing):

	"""
	This class is available through the shortcut :class:`blueprints.Camera <Camera>`.

	Most attribute descriptions are partially taken from `Mujoco <https://mujoco.readthedocs.io/en/latest/XMLreference.html#body-camera>`__.
	
	Attributes
	----------
	fieldofview : int | float, optional
		The angle of field of view in radians.
	ipd : float
		Inter-pupilary distance. This attribute only has an effect during stereoscopic rendering. It 
		specifies the distance between the left and right viewpoints. Each viewpoint is shifted by 
		+/- half of the distance specified here, along the X axis of the camera frame.
	target : blue.Body | None
		The target attribute specifies a Body that is tracked by the camera. Compatible :attr:`mode` values 
		are ('targetbody', 'targetbodycom').
	mode : str
		The possible values are: 
		 * ``'fixed'``: Position and orientation will change w.r.t. the parents frame of reference.
		 * ``'track'``:  Position is at a constant offset from the parent in world coordinates, while the 
		   camera orientation is constant in world coordinates. 
		 * ``'trackcom'``:  is similar to ``'track'`` but the constant spatial offset is defined relative 
		   to the center of mass of the kinematic subtree starting at the parent Body. This can be used to 
		   keep an entire mechanism in view. Note that the subtree center of mass for the World is the 
		   center of mass of the entire model.
		 * ``'targetbody'``: the Camera position is fixed in the parent Body, while the Camera orientation 
		   is adjusted so that it always points towards the targeted Body (which is specified with the 
		   target attribute)
		 * ``'targetbodycom'``: is the same as ``'targetbody'`` but the Camera is oriented towards the 
		   center of mass of the subtree starting at the target Cody.
	"""
	
	@blue.restrict
	def __init__(self, 
		     pos:         np.ndarray|list[int|float] = [0., 0., 0.],
		     alpha:       int|float                  = TAU/4, 
		     beta:        int|float                  = 0., 
		     gamma:       int|float                  = 0., 
		     mode:        str                        = 'fixed', 
		     resolution:  np.ndarray|list[int]       = [480, 480],
		     fieldofview: int|float                  = TAU/8, 
		     ipd:         int|float                  = 0.068, 
		     fps:         int|float|None             = 30, 
		     name:        str|None                   = None, 
		     x:           int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     y:           int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     z:           int|float|np.int32|np.int64|np.float32|np.float64|None = None, 
		     quat:        list[int|float]|np.ndarray|None = None, 
		     **kwargs) -> None:
		"""
		Parameters
		----------
		pos : np.ndarray | list[int | float], optional
			Represents the position of the object. Changing this attribute also changes the properties 
			:attr:`x`, :attr:`y` and :attr:`z`.
		alpha : int | float, optional
			(Improper) euler angle of rotation around the x-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		beta : int | float, optional
			(Improper) euler angle of rotation around the y-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		gamma : int | float, optional
			(Improper) euler angle of rotation around the z-axis in radian. Changing this value also 
			changes the :attr:`euler` property.
		mode : str, optional
			The mode defines who the position and orientation is changed over time. See :attr:`mode` 
			for a detailed description.
		fieldofview : int | float, optional
			The angle of field of view in radians.
		ipd : int | float, optional
			Inter-pupilary distance. This attribute only has an effect during stereoscopic rendering. 
			It specifies the distance between the left and right viewpoints. Each viewpoint is shifted 
			by +/- half of the distance specified here, along the X axis of the camera frame.
		name : str | None, optional
			The user specified name of the Camera. It might change in the case of a naming conflict. 
		x : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the X position coordinate.
		y : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Y position coordinate.
		z : int | float |np.int32 | np.int64 | np.float32 | np.float64 | None, optional
			If `pos` is not specified, this argument sets the Z position coordinate.
		quat: list [ int | float ] | np.ndarray | None, optional
			If set, the quaternion orientation overwrites the euler angles ``alpha``, ``beta`` and ``gamma``.
		**kwargs
			Keyword arguments are passed to ``super().__init__``.
		"""
		assert mode in self._MODES
		self.resolution   = resolution
		self.mode         = mode
		self.fieldofview  = float(fieldofview)
		self.ipd          = ipd
		self.fps          = fps
		super().__init__(pos=pos,
				 alpha=alpha, 
				 beta=beta, 
				 gamma=gamma, 
				 quat=quat, 
				 name=name, 
				 x=x, 
				 y=y, 
				 z=z, 
				 **kwargs)

	@property
	def observation(self):
		"""
		The Camera observation is retrieved as the image rendered by the mujoco Renderer using OpenGL.

		Returns
		-------
		np.ndarray
			RGB image using the specified camera settings.
		"""
		if not isinstance(self.root, blue.WorldType):
			raise Exception(f'Camera observations can only be called once they have been build as part of a World.')
		width, height = self.resolution
		#context = mujoco.GLContext(height, width)
		if not hasattr(self, '_renderer'):
			#self._renderer    = mujoco.Renderer(self.root._mj_model, width, height)
			self._renderer    = mujoco.Renderer(self.root._mj_model, height, width)
			self._last_render = float('-inf')
		if self.root._mj_data.time > self._last_render + self.frameinterval:
			self._renderer.update_scene(self.root._mj_data, self.name)
			for geom in self.root.all.geoms:
				if isinstance(geom, blue.HFieldGeomType):
					mujoco.mjr_uploadHField(self.root._mj_model, self._renderer._mjr_context, geom._index)
			self._rgb = self._renderer.render()
			self._last_render = self.root._mj_data.time
			#print()
			#exit()
		#context.free()
		return self._rgb


	@blue.restrict
	def start_recording(self, 
			    filename: str, 
			    fps:      int = 60, 
			    **kwargs):
		"""
		This shorthand method starts the recording of a video. It can only be 
		used if the camera is already attached within a :class:`World <blueprints.world.World>`.

		For details read :meth:`World.start_recording <blueprints.world.World.start_recording>`.
		
		Parameters
		----------
		filename : str
			The name (with a proper file extension indicating the format) of the video.
			If multiple cameras are passed, multiple videos will be created with the name 
			of each camera prefixed to the filename.
		fps : int
			The frames per second
		**kwargs : dict
			Additional keyword arguments that are passed to imageio to write the video.
		"""
		if isinstance(self.root, blue.WorldType):
			self.root.start_recording(self, 
						  filename=filename, 
						  fps=fps, 
						  **kwargs)
		else:
			raise Exception('A camera can only start recording if it is inhabitant of a World.')

	@blue.restrict
	def stop_recording(self):
		"""
		This shorthand method stops the recording of a video. It can only be 
		used if the camera is already attached within a :class:`World <blueprints.world.World>`.

		For details read :meth:`World.start_recording <blueprints.world.World.start_recording>`.
		"""
		if isinstance(self.root, blue.WorldType):
			self.root.stop_recording(self)
		else:
			raise Exception('A camera can only stop recording if it is inhabitant of a World.')


	@blue.restrict
	def looking(self, 
		    pos:       list[int|float]|np.ndarray|blue.MoveableThingType|None = None, 
		    direction: list[int|float]|np.ndarray|None = None, 
		    **kwargs) -> blue.CameraType:
		"""
		The method creates a new camera that is focussed on pos. The Camera orientation is 
		adjusted to center the ``pos`` argument in the image. Alternatively a direction in 
		which the camera should look can be specified.
		This only constraints two degrees of freedom, so the thrid degree is chosen such that 
		the camera records parallel to the ground.

		Parameters
		----------
		pos : list [ int | float] | np.ndarray | blue.MoveableThingType | None
			If set, it can either be a (global) position or a :class:`MoveableThing <blueprints.thing.moveable.MoveableThing>` 
			that will be centered
		direction :  list [ int | float] | np.ndarray | None
			If set, the returned camera will be oriented parallel to the direction vector.

		Returns
		-------
		blue.CameraType
			A new Camera looking into the specified direction or onto the specified position/Thing
		"""
		if pos is None and direction is None:
			raise ValueError(f'Precisely one argument pos or direction has to be set, but both were None.')
		if pos is not None and direction is not None:
			raise ValueError(f'Precisely one argument pos or direction has to be set, but both were given: pos={pos} direction={direction}')
		if pos is not None:
			if isinstance(pos, list):
				pos = np.array(pos, dtype=np.float32)
			head = self.global_pos
			tail = self._get_pos(pos)
			path =  tail - head
		else:
			path = direction
		# BORROWING MATH FROM TUBE
		length = np.linalg.norm(path)
		# RECONSTRUCT REFERENCE FRAME
		R_z    = path/length
		R_z1, R_z2, R_z3 = R_z
		if R_z1 == R_z2 == 0:
			alpha = 0 if R_z3 > 0 else blue.geometry.PI
			beta  = 0
			gamma = 0
		else:
			R_x = blue.geometry.Vector.normalize(np.array([R_z2, -R_z1, 0]))
			R_y = np.cross(R_z, R_x)
			R   = np.stack([R_x,-R_y,-R_z], axis=1)
			# RECONSTRUCT ANGLES
			alpha, beta, gamma = blue.geometry.Rotation.reference_frame_to_euler(R)
		return self.align(alpha=alpha, 
				  beta=beta, 
				  gamma=gamma, 
				  globally=True, 
				  **kwargs)

	# XML METHODS/FUNCTIONS

	@blue.restrict
	@classmethod
	def _from_xml_element(cls, xml_element) -> blue.ThingType:
		init_args, post_args, rest_args = cls._xml_element_args(xml_element)
		init_args['copy'] = False
		obj = object.__new__(cls)
		obj.__init__(**init_args)
		for key, val in post_args.items():
			setattr(obj, key, val)
		if 'target' in rest_args:
			obj._target_name = rest_args['target']
		return obj


	@blue.restrict
	def _build(self, parent, world, indicies, **kwargs):
		"""
		This method is called to build the xml.

		Parameters
		----------
		parent : xml.etree.ElementTree.Element
			The xml element of its parent
		world : WorldType
			The World from which the build method was called initially

		Returns
		-------
		xml.etree.ElementTree.Element
			The builded xml element of the Thing.
		"""
		if self.target is not None:
			return super()._build(parent, world, indicies, target=self.target.name)
		else:
			return super()._build(parent, world, indicies)

	# KINEMATIC TREE PROPERTIES

	@property
	def target(self) -> blue.BodyType|None:
		"""
		The target attribute specifies a Body that is tracked by the camera. Compatible :attr:`mode` 
		values are ('targetbody', 'targetbodycom').
		
		Returns
		-------
		blue.BodyType | None
		"""
		if hasattr(self, '_target'):
			return self._target
		else:
			return None


	@target.setter
	@blue.restrict
	def target(self, target: blue.BodyType|None) -> None:
		"""
		The target attribute specifies a Body that is tracked by the camera. Compatible :attr:`mode` 
		values are ('targetbody', 'targetbodycom').
		
		Parameters
		----------
		target : blue.BodyType | None
			A body which is tracked by the camera. The Body must be part of the same kinematic tree as 
			he Camera.
		
		Raises
		------
		ValueError
			If the parent is not part of the same kinematic tree or the camera has no parent an error 
			is raised.
		"""
		if self.parent is None and not self._IGNORE_CHECKS:
			raise ValueError(f'To set the target reference, parent must be set first.')
		if self.parent is not None and self.parent is target:
			raise ValueError(f'Camera target cannot be the same as its parent. Use the modes ("fixed", "track" "trackcom") instead.')
		if self.target is not None:
			self.target.detach(self)
		if target is not None:
			target._targeting_cameras.append(self)
		self._target = target

	# XML PROPERTIES

	@property
	def resolution(self) -> np.ndarray:
		"""
		The resolution is used for :attr:`observation` rendering and :meth:`recording <blueprints.world.World.start_recording>`. 
		Rendering in :meth:`World.view <blueprints.world.World.view>` uses OpenGL defaults.

		Returns
		-------
		np.ndarray
			The resolution of the camera.
		"""
		return self._resolution.copy()


	@resolution.setter
	@blue.restrict
	def resolution(self, resolution: np.ndarray|list[int|float]) -> None:
		if isinstance(resolution, np.ndarray) and resolution.shape != (2,):
			raise ValueError(f'Position attribute resolution must have a dimension of 2 got {resolution.shape} instead.')
		if isinstance(resolution, list) and len(resolution) != 2:
			raise ValueError(f'Position attribute resolution must have a dimension of 2 got {len(resolution)} instead.')
		self._resolution = np.array(resolution, dtype=np.uint16)


	@property
	@blue.restrict
	def ipd(self) -> float:
		"""
		This attribute only has an effect during stereoscopic rendering. 
		It specifies the distance between the left and right viewpoints. 
		Each viewpoint is shifted by :math:`\\pm` half of the distance 
		specified here, along the :math:`X` axis of the camera frame.
		
		Returns
		-------
		float
			Inter-pupilary distance
		"""
		
		return self._ipd


	@ipd.setter
	@blue.restrict
	def ipd(self, ipd: int|float) -> None:
		"""
		Parameters
		----------
		ipd : float | int
			Inter-pupilary distance
		"""
		self._ipd = float(ipd)


	@property
	@blue.restrict
	def fieldofview(self) -> float:
		"""
		Vertical field-of-view of the camera.

		Returns
		-------
		float
			Field of view
		"""
		return self._fieldofview


	@fieldofview.setter
	@blue.restrict
	def fieldofview(self, fieldofview: int|float) -> None:
		"""
		Parameters
		----------
		fieldofview : float | int
			Field of view
		"""
		self._fieldofview = float(fieldofview)


	@property
	@blue.restrict
	def fps(self) -> float|None:
		"""
		Frames per second
		
		Returns
		-------
		float
		"""
		return self._fps


	@fps.setter
	@blue.restrict
	def fps(self, fps: int|float|None) -> None:
		"""
		If `fps` is set to `None` a new image will be rendered on each forward step of the physics engine.

		Parameters
		----------
		fps : float | int | None
			Frames per second
		"""
		if fps is not None and fps <= 0.:
			raise ValueError('Frames per second can only be positive.')
		self._fps = float(fps) if fps is not None else None
	
	# DERIVATIONS

	@property
	def fovy(self) -> float:
		"""
		Vertical field-of-view of the camera.
		
		Returns
		-------
		float
			field of view in degrees
		"""
		return self.fieldofview * RADIANS_TO_DEGREES


	@fovy.setter
	@blue.restrict
	def fovy(self, 
		 fovy: float):
		"""
		Parameters
		----------
		fovy : float
			field of view in degrees
		"""
		self.fieldofview = fovy * DEGREES_TO_RADIANS


	@property
	def frameinterval(self) -> float:
		"""
		The delay between each rendering in seconds
		
		Returns
		-------
		float
		"""
		if self.fps is None:
			return 0.
		else:
			return 1/self.fps
